/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __Dasher_H__
#define __Dasher_H__

#include "Geometry.h"
#include "Agent.h"

struct AtomicAction;
class PlayerState;

enum DashType {
	DT_None,
	DT_Forward,     // dash_dir = 0
	DT_Sideward,    // dash_dir = -90 or 90
	DT_Forsideward, // ٲ
	DT_Sideforward  // Ȳ
};

struct DashGetBallInfo
{
	Vector mIntPpos;   //õĵ
	int mBallCycle;   //ballе
	double mIntCycle; //õ
	double mYBuf;      //ֱܵĲԣʱתֱܣԣС¶ȥ뿼ǣ
	bool mIsInverse;     //ܣǷǵ
	DashType mDashType;
	AngleDeg mDashDir;
	int mChangeCycle; //ıdash״̬ʣ
	int mDashPower[4]; //dash״̬תƵǰһڵpower滮ֱΪպõպպò͸պùpower

private:
	friend class Dasher;

	bool mIsSuccess;
};

class Dasher
{
    Dasher();

public:
    ~Dasher();

    static Dasher & instance();

    /** ķʽܵĿ
     * Run to the destination point with the fastest method.
     * \param agent the agent itself.
     * \param pos the destination point.
     * \param buffer point buffer, means the real destinantion is a cycle with the certer 
     * 		of pos and radius of buffer.
     * \param power the intend power to use.
     * \param can_inverse true means can run in the inverse direction of the agent's body.
     * \param turn_first true means that the agent should turn first to adjust the direction.
     * \return true if the action excuted successfully.
     */
    bool GoToPoint(Agent & agent, Vector pos, double buffer  = 0.5, double power = 100.0, bool can_inverse = true, bool turn_first = false);

    /** ķʽܵĿ
     * Run to the destination point with the fastest method.
     * \param agent the agent itself.
     * \param act an atomic action caculated to go to pos for this cycle.
     * \param pos the destination point.
     * \param buffer point buffer, means the real destinantion is a cycle with the certer 
     * 		of pos and radius of buffer.
     * \param power the intend power to use.
     * \param can_inverse true means can run in the inverse direction of the agent's body.
     * \param turn_first true means that the agent should turn first to adjust the direction.
     */
    void GoToPoint(Agent & agent, AtomicAction & act, Vector pos, double buffer  = 0.5, double power = 100.0, bool can_inverse = true, bool turn_first = false);

    /** ȷƣָܺܣܵĿ
     * Run to the destination point with a certain posture, forward or backword.
     * \param agent the agent itself.
     * \param act an atomic action caculated to go to pos for this cycle.
     * \param pos the destination point.
     * \param buffer point buffer, means the real destinantion is a cycle with the center
     * 		of pos and radius of buffer.
     * \param power the intend power to use.
     * \param inverse true means run backward.
     * \param turn_first true means that the agent should turn first to adjust the direction.
     */
    void GoToPointWithCertainPosture(Agent & agent, AtomicAction & act, Vector pos, double buffer  = 0.5, double power = 100.0, const bool inverse = false, bool turn_first = false);

    /**
     * player ܵ target С
     * This function returns the minimum cycles for a player to go to a target position.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param can_inverse true means consider running backwards.
     * @param buffer
     * @return an integer to show the minimum cycles caculated.
     */
    int CycleNeedToPoint(const PlayerState & player, Vector target, bool can_inverse = true, double *buf = 0);

     /**
     * player ܵ target С, ʵ
     * This function returns the minimum cycles for a player to go to a target position.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param can_inverse true means consider running backwards.
     * @return a double to show the minimum cycles caculated.
     */
    double RealCycleNeedToPoint(const PlayerState & player, Vector target, bool can_inverse = true)
    {
    	double buf;
    	int cyc = CycleNeedToPoint(player, target, can_inverse, & buf);
    	return cyc - buf;
    }

    /**
     * player ȷƣָܺܣܵ target Ҫ
     * This function returns the minimum cycles for a player to go to a target position with 
     * a certain posture, forward or backward.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param inverse true means running backwards.
     * @param buffer 
     * @return an integer to show the minimum cycles caculated.
     */
    int CycleNeedToPointWithCertainPosture(const PlayerState & player, Vector target, const bool inverse, double *buf = 0);

    /**
     * player ȷƣָܺܣܵ target Ҫ, ʵ
     * This function returns the minimum cycles for a player to go to a target position with 
     * a certain posture, forward or backward.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param inverse true means running backwards.
     * @return an integer to show the minimum cycles caculated.
     */
    double RealCycleNeedToPointWithCertainPosture(const PlayerState & player, Vector target, const bool inverse)
    {
    	double buf;
    	int cyc = CycleNeedToPointWithCertainPosture(player, target, inverse, & buf);
    	return cyc - buf;
    }

    /**
     * ͬһdash_powerĳҪʱ
     * Caculate cycles needed to a target position with a certain dash power.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param dash_power the power for dash to predict.
     * @return an integer to show the cycles caculated.
     */
    int CyclePredictedToPoint(const PlayerState& player , Vector target , double dash_power);

    /**
     * Աĳʼ,ٵӳ
     * Caculate the delay cycles for a player to accel to his max speed.
     * @param player the player to caculate.
     * @param angle the max speed angle to predict.
     * @return an double to show the cycles caculated.
     */
    double CalcPlayerCycleFix(const PlayerState &player, AngleDeg angle);

    /**
     * ĳԱܵĳڣpre(1)λ㣬Ȼ+1
     * Caculate the cycles needed for a player to go to the point.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @return an double to show the cycles caculated.
     */
    double CalcPlayerGoToPointCycle(const PlayerState & player, const Vector & target);

    /**
     * ҪΪ˵powerãΪ˱powerܹ
     * Ŀ㡣ԭΪ˸߲ЭpowerֻܼС
     * Ϊ˲ҪdashpowerҲɿΪ㡣Ҳ
     * dashִ
     * power 
     * This funcition is mainly used to adjust ( usually decrease ) the dash power 
     * to avoid run over the target position by using a big power. The final power 
     * could be 0 to avoid an unnecessary dash.
     * The sign of power is given outside the function.
     * @param player the player to caculate.
     * @param target the target position to go to.
     * @param buffer
     * @param power orignal power given.
     * @return power adjusted.
     */
    double AdjustPowerForDash(const PlayerState & player, Vector target, double buffer, double power);

    /**
     * תֱdash
     * Planing to turn or to dash.
     * @param agent the agent itself.
     * @param act an atomic action to return the caculated decision.
     * @param target the target position to go to.
     * @param buffer
     * @param power the intend power for dash.
     * @param inverse true means running backwards.
     * @param turn_first true means turn first manually.
     */
    void TurnDashPlaning(Agent & agent, AtomicAction & act, Vector target, double buffer, double power, bool inverse, bool turn_first = false);

    /**
     * Caculate the get ball cycles and actions.
     * @param agent the agent itself.
     * @param act the atomic action to execute this cycle to get the ball.
     * @param int_cycle the beginning cycle to do the action, while -1 means now.
     * @param can_inverse true means allow runnning backwards to get the ball.
     * @param turn_first true means the agent turn first to get the ball.
     * @return a double to show the cycles needed and if it less than 0, that is to say impossible.
     */
    double GetBall(Agent & agent, AtomicAction & act, int int_cycle = -1, bool can_inverse = true, bool turn_first = false);

    /**
     * Caculate the get ball action and execute it.
     * @param agent the agent itself.
     * @param int_cycle the beginning cycle to do the action, while -1 means now.
     * @param can_inverse true means allow runnning backwards to get the ball.
     * @param turn_first true means the agent turn first to get the ball.
     * @return true means the action executed successfully.
     */
    bool GetBall(Agent & agent, int int_cycle = -1, bool can_inverse = true, bool turn_first = false);

    /**
     * ϸֱǷõ
     * Caculate in detail if the agent can get the ball only by dashing.
     * @param agent the agent itself.
     * @param cycle cycles forward to caculate.
     * @return true means the agent can get the ball.
     */
    bool CalcDashGetBallInfo(Agent & agent, int cycle = 10);

    /*
     * This function is used to correct the target position when near the ball.
     * @param player the player to consider.
     * @param target the target position to go to.
     * @param fix the distance error to fix.
     * @return position corrected.
     */
    Vector CorrectTurnForDash(const PlayerState & player, const Vector & target, double fix = 0.0);

public:
	//ģʽȦƽ
	//the funcitions below are for move around the forbidden circle or penalty area.
	
        /*
	 * Fix the route when needed by using the next 3 funcions.
	 * @param agent the agent itself.
	 * @param origin the origin position to go to.
	 * @return new target position fixed from the origin one.
	 */
	Vector FixRoute(const Agent &agent, Vector origin);

	/*
	 * Avoid the forbidden circle when the opponent team taking a free kick.
	 * @param agent the agent itself.
	 * @param origin the origin position to go to.
	 * @return new target position fixed from the origin one.
	 */
	Vector AvoidCircle(const Agent &agent, Vector origin);

        /*
	 * Avoid the penalty area when the opponent team taking a goal kick.
	 * @param agent the agent itself.
	 * @param origin the origin position to go to.
	 * @return new target position fixed from the origin one.
	 */
	Vector AvoidPenaltyArea(const Agent &agent, Vector origin);

	/*
	 * Caculate a contact line of the forbidden circle to go along when avoiding the circle.
	 * @param agent the agent itself.
	 * @param radius the radius of the cycle.
	 * @param project_point the intend position to go to.
	 * @return new target position fixed from the origin one.
	 */
	Vector GoContactLine(const Agent &agent, double radius, Vector project_point);

private:
	/**
	* player ֻdashܵ target С
	* Ϊֱͨdashܵĳʱܽ׶κͲܽ׶ûн
	* @param player
	* @param target
	* @param dashpower 4ά,dashٶ,ڵĻֱܵdashpower,ܵ,ܹ,ܹ
	* @param buf ԣ,簴ܵʱ0.5,0.5 ԵʱԱٶȵõbuf
	* @return
	*/
	/*
	 * Caculate the minimum cycles for a player dashing to a point considering that the dash method
	 * is forward/backward and then sideward or sideward and then forward/backward.
	 * @param player the player to caculate.
	 * @param expected_cycle the expected cycles for the player to dash to the point.
	 * @param target the target position to go to.
	 * @param dash_dir the direction of dash.
	 * @param dash_type the type of dash.
	 * @param change_cycle the cycle change caculated.
	 * @param buf cycle allowance.
	 * @param dashpower an array of 4 units, stores 4 dash power for 
	 * 	1.dash to get the ball
	 * 	2.dash to collide with ball
	 * 	3.dash to avoid collision
	 * 	4.dash to miss ball.
	 * @return an integer to show the cycles needed.
	 */
	int CycleNeedToDashToPoint(const PlayerState & player, int expected_cycle, const Vector target, AngleDeg & dash_dir, DashType & dash_type, int & change_cycle, double buf[2], int *dashpower);

	/**
	* dash dist  
	* @param dist dashľ
	* @param speed Ŀ귽ϵĳٶ
	* @param accrate ٶʣdir_rateģ
	* @param double dir_rate ֱϵ
	* @param dashpower 4ά,dashٶ,ڵĻֱܵdashpower,ܵ,ܹ,ܹ -- ﶼģinverseܣҪȡ
	* @param buf ԣ,簴ܵʱ0.5,0.5 ԵʱԱٶȵõbuf
	* @return
	*/
	/*
	 * Caculate the cycles for a player dashing for a distance.
	 * @param player the player to caculate.
	 * @param expected_cycle the expected cycles for player to dash.
	 * @param speed expected speed for player to dash.
	 * @param decay the player's decay.
	 * @param accrate the player's accel rate (considered the direction rate).
	 * @param vertical_dist the vertical distance (left or right) for player to dash.
	 * @param kick_area the player's kick area.
	 * @param stamina the player's stamina.
	 * @param stamina_recovery_thr the player's stamina recovery threshold.
	 * @param stacostrate the stamina cost rate for the dash command. 
	 * @param stamina_inc_max the player's max stamina increase per cycle.
	 * @param buf cycle allwoance.
	 * @param dashpower the power for dash.
	 * @return an integer to show the cycles needed.
	 */
	int CycleNeedToDashForDist(double dist, const int expected_cycle, double speed, const double decay, const double accrate, const double vertical_dist, const double kick_area, double stamina, const double stamina_recovery_thr, const double stacostrate, const double stamina_inc_max, double *buf = 0, int *dashpower = 0);

	//for debug
	void DebugCycleNeedToDashForDist(const Vector start, const Vector target, double dist, const int expected_cycle, double speed, const double decay, const double accrate, const double vertical_dist, const double kick_area, double stamina, const double stamina_recovery_thr, const double stacostrate, const double stamina_inc_max);

	//only 10 cycles considered, and very low ball speed is out of consideration.
    DashGetBallInfo mDashGetBallInfo[10]; //dashֻǶ,ٹС

public:
	/*
	 * õ cycle ڵdashgetballinfo
	 * Get the dashgetballinfo for a certain cycle.
	 * @param cycle the cycle of the dashgetballinfo to get.
	 * @return the dashgetballinfo of the cycle.
	 */
    const DashGetBallInfo & GetDashGetBallInfo(const int cycle) const { Assert(cycle >= 1 && cycle <= 10); return mDashGetBallInfo[cycle - 1]; }

private:
	AgentID mDashGetInfoAgentID;

//==============================================================================
// ṩйagentΪĽӿڣֱӷAtomicActionⲿ
// The following funtions are some interfaces for agent to do some basic behaviors 
// without ball. The return value are all AtomicAction. 
public:
    /**
     * ײ
     * Collide with the ball.
     * @param agnet the agent itself.
     * @return an atomic action to collide with the ball.
     */
    AtomicAction GetCollideBallAction(const Agent &agent);

    /**
     * תض
     * Turn body to a certain direction.
     * @param agent the agent itself.
     * @param ang the angle to turn to.
     * @return an atomic action to turn the body.
     */
    AtomicAction GetTurnBodyToAngleAction(const Agent & agent, AngleDeg ang);

    /*! This method determines the optimal dash power to mantain an optimal speed
	When the current speed is too high and the distance is very small, a
	negative dash is performed. Otherwise the difference with the maximal speed
	is determined and the dash power rate is set to compensate for this
	difference.
	\param posRelTo relative point to which we want to dash
	\param angBody body angle of the agent
	\param vel current velocity of the agent
	\param dEffort current effort of the player
	\param iCycles desired number of cycles to reach this point
	\return dash power that should be sent with dash command */
    bool GetPowerForForwardDash(const Agent &agent, double* dash_power, Vector posRelTo, double angBody, double dEffort, int iCycles );

public:
    /**
     * ȷ㳯ĳܶԶ(kickareaܹԶ),ãһkick
     * @dir 
     * @maxdis ࿼ǶԶ
     * @dis ַdis¼ڸǰľ룬kick cyc,turn cycΪdis Ϊ0,ҲdisǴӿʼdash¼
     * @return ˶cyc,dis[cyc]=maxdis,dis[0]kickǸ
     */
    /*
     * Caculate in detail how far a player can rich considering the player's kick area
     * supposing that in the first cycle the player do a kick. This function is for dribble.
     * @param player_state the player state for the player.
     * @param dir the dash direction.
     * @param maxdis the maximum distance to caculate.
     * @param maxcyc the maximum cycles to caculate.
     * @param dis the pointer for an array of double to record the distance for each cycle. 
     * 		The kick cycle or turn cycle are considered dis[0].
     * @return an integer show the cycles. dis[cyc]=maxdis.
     */
    int CalcDashDist(const PlayerState & player_state, const double & dir, const double & maxdis, const int & maxcyc, double * dis);


    /**
	 * ĳԱǰܶԶdash
     *
     * @player_state ָԱ
     * @max_cycle ǵ
     * @distance ÿǰľ
     * @speed ÿڵٶȴС
     * @stamina ÿԱ
     * @is_fist_kick һǷ
     */
    /*
     * Caculate how far a player can rich by dashing with a power above 0.
     * @param player_state the player state for the player.
     * @param max_cycle the maximum cycles to caculate.
     * @param distance an array of double to store the distance of each cycle.
     * @param speed an array of double to store the speed of each cycle.
     * @param stamina an array of double to store the stamina of the player of each cycyle.
     * @param is_first_kick true means in the first cycle the player does a kick.
     */
    void CalcFrontDashDist(const PlayerState & player_state, const int & max_cycle, double * distance, double * speed, double * stamina, bool is_first_kick = true);

    /**
     * dashںλüٶ
     *
     * desiny,ballpathΪתƽƵԼΪԲ꣬balloutSpeed0ΪԵ
     */
    /*
     * Caculate the relative position of ball and the speed for ball to kick to.
     * @param ball_state the ball state.
     * @param player_state the player state.
     * @param max_cycle the max_cycle to consider.
     * @param denstiny the destination position ( relative coordination ).
     * @param ball_path an array to store the ball path ( relative coordination ) every cycle.
     * @param ballOutSpeed0 the fist speed for ball.
     */
    void CalcBallPath(const BallState & ball_state, const PlayerState & player_state, const int & max_cycle, Vector & destiny, Vector * ball_path, Vector & ballOutSpeed0);

    /**
     * Աdirܣÿڵľ
     * @param player_state
     * @param dir Ŀ귽
     * @param dis 
     * @param maxdis 
     * @param maxcyc 
     * @param adjust_finish ʱǶȵ
     * @return
     */
    /*
     * Caculate the distance for a player to run to a certain direction per cycle.
     * @param player_state the player state.
     * @param dir the target direction.
     * @param dis an array to record the distance per cycle.
     * @param maxdis the maximum distance to caculate.
     * @param maxcyc the maximum cycles to caculate.
     * @param adjust_finish the time for turning to the right direction finished.
     * @param in_future true means to caculate a player's future state.
     * @return an integer to return the number of cycyles caculated.
     */
    int CalcRunDistance(const PlayerState & player_state, double dir, double *dis, double maxdis, int maxcyc, int & adjust_finish, bool in_future);

public:
	static double GETBALL_BUFFER; //ʹõжǷߵbufworldstateĴ
};

inline int Dasher::CycleNeedToDashForDist(double dist, const int expected_cycle, double speed, const double decay, const double accrate, const double vertical_dist, const double kick_area, double stamina, const double stamina_recovery_thr, const double stacostrate, const double stamina_inc_max, double *buf, int *dashpower)
{
	Assert(vertical_dist > -FLOAT_EPS);

	if (vertical_dist > kick_area) return 1000;

	const double d1 = sqrt( Sqr( kick_area ) - Sqr( vertical_dist ) );
	const int DashMax = (int)ServerParam::instance().maxDashPower();
	const double pSize = ServerParam::instance().playerSize(); //player_size

	int cycle = 0;
	while(cycle < 16){
		++cycle;
		if (cycle > expected_cycle) return 1000;
		double dDash = int(MinMax( stamina_inc_max ,( stamina - stamina_recovery_thr + stamina_inc_max ) / 2, DashMax ) ); //dashֵ
		double DPower = ( dist - speed - d1 ) / (accrate);
		if(DPower <= dDash){//ܵ,ϸ
			if (dashpower) {
				dashpower[0] = (int)DPower; //պõ
				if(pSize > fabs(vertical_dist)){
					double d2 = sqrt( Sqr( pSize ) - Sqr( vertical_dist ) );
					dashpower[1] = (int)(( dist - speed - d2 ) / accrate) ; //պ
					dashpower[2] = (int)(( dist - speed + d2 ) / accrate); //պò
				}
				else{
					dashpower[1] = 1000;//unvalid
					dashpower[2] = 1000;//unvalid
				}
				dashpower[3] = (int)(( dist - speed + d1 ) / accrate); //պù
			}
			//Ѿbufתcycle
			if (buf) {
				buf[0] = ( dDash - DPower ) * accrate; //ʱspeedǸ,bufǸʮ,
				buf[0] = MinMax(0.0, *buf , 0.99);
				buf[1] = fabs(( kick_area - vertical_dist));
			}
			return cycle;
		}
		stamina -= dDash * stacostrate - stamina_inc_max;
		speed += dDash * accrate;
		dist -= speed;
		speed *= decay;
	}

	return 1000;
}

#endif
