/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef DECISIONTREE_H_
#define DECISIONTREE_H_

#include <list>
#include "BehaviorBase.h"

class Agent;

class DecisionTree {
public:
	DecisionTree() {}
	virtual ~DecisionTree() {}

	/**
	 * 
	 * @param agent
	 */
	/**
	 * Do the decision.
	 * @param agent the agent itself.
	 * @return true means the decision has been made.
	 */
	bool Decision(Agent & agent);

private:
	/**
	* ɶԣ״̬
	* @param agent ǰڵľֲ
	* @param step ǰڵȣǷأstep = 1һ㣬step = max_step ǵһ㣩
	* @return
	*/
	/**
	 * Search the decision tree, and finish the evaluation of conditions and actions.
	 * @param agent the agent itslef.
	 * @param step the depth of current node. step = 1 means the last ply, while step = max_step 
	 * 	means the first ply.
	 * @return an active behavior, the best result of the searching.
	 */
	ActiveBehavior Search(Agent & agent, int step);

	/**
	 * Get the best active behavior from a active behavior list, and store the list for next cycle.
	 * @param agent teh agent itself.
	 * @param behavior_list the list of active behaviors to be sorted and stored.
	 * @return an active behavior which is the best one of the whole list.
	 */
	ActiveBehavior GetBestActiveBehavior(Agent & agent, std::list<ActiveBehavior> & behavior_list);

	/**
	 * Mutex plan function for variaties of behaviors.
	 * @param agent the agent itself.
	 * @param active_behavior_list the list of active behaviors as the result of planning.
	 * @return true if the planning gets any active behaviors so that the list is not empty.
	 */
	template <typename BehaviorDerived>
	bool MutexPlan(Agent & agent, std::list<ActiveBehavior> & active_behavior_list){
		BehaviorDerived(agent).Plan(active_behavior_list);
		return !active_behavior_list.empty();
	}
};

#endif /* DECISIONTREE_H_ */
