/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __PositionInfo_H__
#define __PositionInfo_H__

#include <cstdlib>
#include <vector>
#include "InfoState.h"

class WorldState;

/**
* йسԱλõInfo
*/
/**
 * Position Information
 */
class PositionInfo: public InfoStateBase
{
public:
	PositionInfo(WorldState *pWorldState, InfoState *pInfoState);
	virtual ~PositionInfo();

	/** Dist Information */
	const double & GetBallDistToPlayer(Unum unum) const;
	const double & GetPlayerDistToPlayer(Unum unum1, Unum unum2) const;
	const double & GetBallDistToTeammate(Unum unum) const { Assert(unum >= 1 && unum <= TEAMSIZE); return GetBallDistToPlayer(unum); }
    const double & GetBallDistToOpponent(Unum unum) const { Assert(unum >= 1 && unum <= TEAMSIZE); return GetBallDistToPlayer(-unum); }

	/** Player List Sort by X */
    const std::list<KeyPlayerInfo> & GetXSortTeammate();
    const std::list<KeyPlayerInfo> & GetXSortOpponent();

	/** Close Information */
	/** Close to Point */
	std::vector<Unum> GetClosePlayerToPoint(const Vector & bp, const Unum & unum = 0) const;

	/** Close to Ball */
	const std::vector<Unum> & GetClosePlayerToBall();
	const std::vector<Unum> & GetCloseTeammateToBall();
	const std::vector<Unum> & GetCloseOpponentToBall();

	/** Close to Player */
	const std::vector<Unum> & GetClosePlayerToPlayer(Unum i);
	const std::vector<Unum> & GetCloseTeammateToPlayer(Unum i);
	const std::vector<Unum> & GetCloseOpponentToPlayer(Unum i);

	/** Close to Teammate */
	const std::vector<Unum> & GetClosePlayerToTeammate(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetClosePlayerToPlayer(i); }
	const std::vector<Unum> & GetCloseTeammateToTeammate(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetCloseTeammateToPlayer(i); }
	const std::vector<Unum> & GetCloseOpponentToTeammate(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetCloseOpponentToPlayer(i); }

	/** Close to Opponent */
	const std::vector<Unum> & GetClosePlayerToOpponent(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetClosePlayerToPlayer(-i); }
	const std::vector<Unum> & GetCloseTeammateToOpponent(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetCloseTeammateToPlayer(-i); }
	const std::vector<Unum> & GetCloseOpponentToOpponent(Unum i) { Assert(i >= 1 && i <= TEAMSIZE); return GetCloseOpponentToPlayer(-i); }

	/** Closest to Ball */
	Unum GetClosestPlayerToBall()   { return GetClosePlayerToBall().empty()? 0: GetClosePlayerToBall()[0]; }
	Unum GetClosestTeammateToBall() { return GetCloseTeammateToBall().empty()? 0: GetCloseTeammateToBall()[0]; }
	Unum GetClosestOpponentToBall() { return GetCloseOpponentToBall().empty()? 0: GetCloseOpponentToBall()[0]; }

	/** Closest to Player */
	Unum GetClosestPlayerToPlayer(Unum i) { return GetClosePlayerToPlayer(i).empty()? 0: GetClosePlayerToPlayer(i)[0]; }
	Unum GetClosestTeammateToPlayer(Unum i) { return GetCloseTeammateToPlayer(i).empty()? 0: GetCloseTeammateToPlayer(i)[0]; }
	Unum GetClosestOpponentToPlayer(Unum i) { return GetCloseOpponentToPlayer(i).empty()? 0: GetCloseOpponentToPlayer(i)[0]; }

	/** Closest to Teammate */
	Unum GetClosestPlayerToTeammate(Unum i) { return GetClosePlayerToTeammate(i).empty()? 0: GetClosePlayerToTeammate(i)[0]; }
	Unum GetClosestTeammateToTeammate(Unum i) { return GetCloseTeammateToTeammate(i).empty()? 0: GetCloseTeammateToTeammate(i)[0]; }
	Unum GetClosestOpponentToTeammate(Unum i) { return GetCloseOpponentToTeammate(i).empty()? 0: GetCloseOpponentToTeammate(i)[0]; }

	/** Closest to Opponent */
	Unum GetClosestPlayerToOpponent(Unum i) { return GetClosePlayerToOpponent(i).empty()? 0: GetClosePlayerToOpponent(i)[0]; }
	Unum GetClosestTeammateToOpponent(Unum i) { return GetCloseTeammateToOpponent(i).empty()? 0: GetCloseTeammateToOpponent(i)[0]; }
	Unum GetClosestOpponentToOpponent(Unum i) { return GetCloseOpponentToOpponent(i).empty()? 0: GetCloseOpponentToOpponent(i)[0]; }

	/** Closest Dist to Ball */
	double GetClosestTeammateDistToBall() { return GetClosestTeammateToBall() == 0? HUGE_DIST: mpWorldState->GetTeammate(GetClosestTeammateToBall()).GetPos().Dist(mpWorldState->GetBall().GetPos()); }
	double GetClosestOpponentDistToBall() { return GetClosestOpponentToBall() == 0? HUGE_DIST: mpWorldState->GetOpponent(GetClosestOpponentToBall()).GetPos().Dist(mpWorldState->GetBall().GetPos()); }

	/** Off Side Line Information */
	const Unum & GetTeammateOffsideLineOpp() const     { return mTeammateOffsideLineOpp; }
	const double & GetTeammateOffsideLine() const       { return mTeammateOffsideLine; }
	const double & GetTeammateOffsideLineConf() const   { return mTeammateOffsideLineConf; }
	const double & GetTeammateOffsideLineSpeed() const  { return mTeammateOffsideLineSpeed; }

	const Unum & GetOpponentOffsideLineTm() const     { return mOpponentOffsideLineTm; }
	const double & GetOpponentOffsideLine() const       { return mOpponentOffsideLine; }
	const double & GetOpponentOffsideLineConf() const   { return mOpponentOffsideLineConf; }
	const double & GetOpponentOffsideLineSpeed() const  { return mOpponentOffsideLineSpeed; }

	/** õǰߵԱб */
	/** Get Kickable Players List */
	const std::vector<Unum> & GetPlayerWithBallList();

	Unum GetPlayerWithBall() { return GetPlayerWithBallList().empty()? 0: GetPlayerWithBallList()[0]; }
    /**
     * Get closest player in a cycle which radius = buffer;
     * If buffer is larger than 2 then the result will be calculated as buffer = 2;
     * @param buffer;
     * return Unum of player. Unum > 0 ? teammate:opponent/unknow.
     */
    Unum GetPlayerWithBall(const double buffer);

	Unum GetTeammateWithBall();
    /**
     * Get closest teammate in a cycle which radius = buffer;
     * If buffer is larger than 2 then the result will be calculated as buffer = 2;
     * @param buffer;
     * return Unum of teammate.
     */
    Unum GetTeammateWithBall(const double buffer);

	Unum GetOpponentWithBall();
    /**
     * Get closest opponent in a cycle which radius = buffer;
     * If buffer is larger than 2 then the result will be calculated as buffer = 2;
     * @param buffer;
     * return Unum of opponent.
     */
    Unum GetOpponentWithBall(const double buffer);

	OrderedSoftQueue<KeyPlayerInfo, TEAMSIZE> & GetTeammateAnglesToBallQueue();
	OrderedSoftQueue<KeyPlayerInfo, TEAMSIZE> & GetOpponentAnglesToBallQueue();

private:
	/** º */
	/** Update */
	void UpdateRoutine();

	/** ±Ա֮໥תԱʾѣʾ */
	/** transformational relation between array index and player's number. > 0 : teammate ; < 0 : opponent */
	int Unum2Index(Unum unum) const {
		Assert(abs(unum) >= 1 && abs(unum) <= TEAMSIZE);
		return unum > 0? unum: TEAMSIZE - unum;
	}

	int Index2Unum(int index) const {
		Assert(index >= 1 && index <= TEAMSIZE * 2);
		return index <= TEAMSIZE? index: TEAMSIZE - index;
	}

	void UpdateDistMatrix();
	void UpdateOffsideLine();
    void UpdateOppGoalInfo(); /** ʱԺҪٸ */

private:
	double mDistMatrix[1 + 2 * TEAMSIZE][1 + 2 * TEAMSIZE]; // 22Ա໥֮ľ룬0Ϊ1-11Ϊѣ1222Ϊ
															// Dist Matrix of players' and ball's distance to each other. 0: ball; 1~11: teammate; 12~22: opponent;

    std::list<KeyPlayerInfo> mXSortTeammateList;
    std::list<KeyPlayerInfo> mXSortOpponentList;

	std::vector<Unum> mPlayer2BallList;
	std::vector<Unum> mTeammate2BallList;
	std::vector<Unum> mOpponent2BallList;

	std::vector<Unum> mPlayer2PlayerList[1 + 2 * TEAMSIZE];
	std::vector<Unum> mTeammate2PlayerList[1 + 2 * TEAMSIZE];
	std::vector<Unum> mOpponent2PlayerList[1 + 2 * TEAMSIZE];

	Unum mTeammateOffsideLineOpp;
	double mTeammateOffsideLine;
	double mTeammateOffsideLineConf;
	double mTeammateOffsideLineSpeed;

	Unum mOpponentOffsideLineTm;
	double mOpponentOffsideLine;
	double mOpponentOffsideLineConf;
	double mOpponentOffsideLineSpeed;

	std::vector<Unum> mPlayerWithBallList;	//ǰĶԱ -- bufferж
											//kickable player list -- no buffer
	Time mPlayerWithBallList_UpdateTime;

	OrderedSoftQueue<KeyPlayerInfo, TEAMSIZE> mTeammateAnglesToBallQueue;
	OrderedSoftQueue<KeyPlayerInfo, TEAMSIZE> mOpponentAnglesToBallQueue;

private:
	class PlayerDistCompare {
	public:
		bool operator()(const std::pair<Unum, double> & i, const std::pair<Unum, double> & j){
			return i.second < j.second;
		}
	};

public:
	/** The following 3 functions compute the angle seen in ball's view, the angle is global */
    AngleDeg GetPlayerAngle2Ball(Unum unum) const
    {
        return (unum > 0) ? GetTeammateAngle2Ball(unum) : GetOpponentAngle2Ball(-unum);
    }

    AngleDeg GetTeammateAngle2Ball(Unum unum) const
    {
        Assert (unum >= 1 && unum <= TEAMSIZE);
        return mTeammateAngle2Ball[unum - 1];
    }

    AngleDeg GetOpponentAngle2Ball(Unum unum) const
    {
        Assert (unum >= 1 && unum <= TEAMSIZE);
        return mOpponentAngle2Ball[unum - 1];
    }

private:
    AngleDeg mTeammateAngle2Ball[TEAMSIZE]; // ĽǶ
    AngleDeg mOpponentAngle2Ball[TEAMSIZE]; // ĽǶ

public:
    const Vector & GetOppGoal2Ball() const      { return mOppGoal2Ball; }
    const Vector & GetOppLeftPost2Ball() const  { return mOppLeftPost2Ball; }
    const Vector & GetOppRightPost2Ball() const { return mOppRightPost2Ball; }
    const AngleDeg & GetOppGoal2BallAngle() const       { return mOppGoal2BallAngle; }
    const AngleDeg & GetOppLeftPost2BallAngle() const   { return mOppLeftPost2BallAngle; }
    const AngleDeg & GetOppRightPost2BallAngle() const  { return mOppRightPost2BallAngle; }

private:
    Vector mOppGoal2Ball;       // 
								// Mid of the goal
    Vector mOppLeftPost2Ball;   // 
								// Left post
	Vector mOppRightPost2Ball;  // 
								// Right post
	AngleDeg mOppGoal2BallAngle;
    AngleDeg mOppLeftPost2BallAngle;
	AngleDeg mOppRightPost2BallAngle;

public:
    bool IsConsiderAngle(const Unum &self_unum, const AngleDeg &ang, bool consider_dist = false);
};

#endif
